/* Copyright (c) 2006, Nokia Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * * Neither the name of the Nokia Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <glib/gi18n.h>
#include <gtk/gtktreeviewcolumn.h>
#include <modest-runtime.h>

#include <widgets/modest-main-window.h>
#include <widgets/modest-window-priv.h>
#include <widgets/modest-msg-edit-window.h>
#include "modest-account-view-window.h"


#include "modest-widget-memory.h"
#include "modest-ui-actions.h"
#include "modest-main-window-ui.h"
#include "modest-account-mgr.h"
#include "modest-conf.h"
#include <modest-tny-msg.h>
#include "modest-mail-operation.h"
#include "modest-icon-names.h"

/* 'private'/'protected' functions */
static void modest_main_window_class_init    (ModestMainWindowClass *klass);
static void modest_main_window_init          (ModestMainWindow *obj);
static void modest_main_window_finalize      (GObject *obj);

static void restore_sizes (ModestMainWindow *self);
static void save_sizes (ModestMainWindow *self);

static gboolean     on_header_view_button_press_event   (ModestHeaderView *header_view,
							 GdkEventButton   *event,
							 ModestMainWindow *self);

static gboolean     on_folder_view_button_press_event   (ModestFolderView *folder_view,
							 GdkEventButton   *event,
							 ModestMainWindow *self);

static gboolean     show_context_popup_menu             (ModestMainWindow *window,
							 GtkTreeView      *tree_view,
							 GdkEventButton   *event,
							 GtkWidget        *menu);

static void         connect_signals                      (ModestMainWindow *self);


/* list my signals */
enum {
	/* MY_SIGNAL_1, */
	/* MY_SIGNAL_2, */
	LAST_SIGNAL
};

typedef struct _ModestMainWindowPrivate ModestMainWindowPrivate;
struct _ModestMainWindowPrivate {

	GtkWidget *folder_paned;
	GtkWidget *msg_paned;
	GtkWidget *main_paned;
	
	GtkWidget *online_toggle;
	GtkWidget *folder_info_label;
};


#define MODEST_MAIN_WINDOW_GET_PRIVATE(o)      (G_TYPE_INSTANCE_GET_PRIVATE((o), \
                                                MODEST_TYPE_MAIN_WINDOW, \
                                                ModestMainWindowPrivate))

typedef struct _GetMsgAsyncHelper {
	ModestMainWindowPrivate *main_window_private;
	guint action;
	ModestMailOperationReplyType reply_type;
	ModestMailOperationForwardType forward_type;
	gchar *from;
	TnyIterator *iter;
} GetMsgAsyncHelper;

/* globals */
static GtkWindowClass *parent_class = NULL;

/* uncomment the following if you have defined any signals */
/* static guint signals[LAST_SIGNAL] = {0}; */

GType
modest_main_window_get_type (void)
{
	static GType my_type = 0;
	if (!my_type) {
		static const GTypeInfo my_info = {
			sizeof(ModestMainWindowClass),
			NULL,		/* base init */
			NULL,		/* base finalize */
			(GClassInitFunc) modest_main_window_class_init,
			NULL,		/* class finalize */
			NULL,		/* class data */
			sizeof(ModestMainWindow),
			1,		/* n_preallocs */
			(GInstanceInitFunc) modest_main_window_init,
			NULL
		};
		my_type = g_type_register_static (MODEST_TYPE_WINDOW,
		                                  "ModestMainWindow",
		                                  &my_info, 0);
	}
	return my_type;
}

static void
modest_main_window_class_init (ModestMainWindowClass *klass)
{
	GObjectClass *gobject_class;
	gobject_class = (GObjectClass*) klass;

	parent_class            = g_type_class_peek_parent (klass);
	gobject_class->finalize = modest_main_window_finalize;

	g_type_class_add_private (gobject_class, sizeof(ModestMainWindowPrivate));
}

static void
modest_main_window_init (ModestMainWindow *obj)
{
	ModestMainWindowPrivate *priv;
	TnyFolderStoreQuery     *query;
	
	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(obj);
	
	priv->folder_paned = NULL;
	priv->msg_paned    = NULL;
	priv->main_paned   = NULL;	

	/* folder view */
	query = tny_folder_store_query_new ();
	tny_folder_store_query_add_item (query, NULL,
					 TNY_FOLDER_STORE_QUERY_OPTION_SUBSCRIBED);

	obj->folder_view =
		MODEST_FOLDER_VIEW(modest_folder_view_new (modest_runtime_get_account_store(),
							   query));
	if (!obj->folder_view)
		g_printerr ("modest: cannot instantiate folder view\n");	
	g_object_unref (G_OBJECT (query));

	/* header view */
	obj->header_view  =
		MODEST_HEADER_VIEW(modest_header_view_new (NULL, MODEST_HEADER_VIEW_STYLE_DETAILS));
	if (!obj->header_view)
		g_printerr ("modest: cannot instantiate header view\n");

	/* msg preview */
	obj->msg_preview = MODEST_MSG_VIEW(modest_msg_view_new (NULL));
	if (!obj->msg_preview)
		g_printerr ("modest: cannot instantiate msgpreiew\n");

	/* online/offline combo */
	priv->online_toggle = gtk_toggle_button_new ();

	/* label with number of items, unread items for 
	   the current folder */
	priv->folder_info_label = gtk_label_new (NULL);

	/* status bar */
	obj->status_bar   = gtk_statusbar_new ();
        gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR(obj->status_bar),
                                           FALSE);

	/* progress bar */
	obj->progress_bar = gtk_progress_bar_new ();
	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR(obj->progress_bar), 1.0);
	gtk_progress_bar_set_ellipsize (GTK_PROGRESS_BAR(obj->progress_bar),
					PANGO_ELLIPSIZE_END);
}

static void
modest_main_window_finalize (GObject *obj)
{
	G_OBJECT_CLASS(parent_class)->finalize (obj);
}


static void
restore_sizes (ModestMainWindow *self)
{
	ModestConf *conf;
	ModestMainWindowPrivate *priv;
	ModestWindowPrivate *parent_priv;
	
	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(self);
	parent_priv = MODEST_WINDOW_GET_PRIVATE(self);

	conf = modest_runtime_get_conf ();
	
	modest_widget_memory_restore (conf, G_OBJECT(priv->folder_paned),
				      "modest-folder-paned");
	modest_widget_memory_restore (conf, G_OBJECT(priv->msg_paned),
				      "modest-msg-paned");
	modest_widget_memory_restore (conf, G_OBJECT(priv->main_paned),
				      "modest-main-paned");
	modest_widget_memory_restore (conf, G_OBJECT(self->header_view),"header-view");
	modest_widget_memory_restore (conf,G_OBJECT(self), "modest-main-window");
}


static void
save_sizes (ModestMainWindow *self)
{
	ModestWindowPrivate *parent_priv;
	ModestMainWindowPrivate *priv;
	ModestConf *conf;
	
	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(self);
	parent_priv = MODEST_WINDOW_GET_PRIVATE(self);

	conf = modest_runtime_get_conf ();
	
	modest_widget_memory_save (conf,G_OBJECT(self), "modest-main-window");
	modest_widget_memory_save (conf, G_OBJECT(priv->folder_paned),
				   "modest-folder-paned");
	modest_widget_memory_save (conf, G_OBJECT(priv->msg_paned),
				   "modest-msg-paned");
	modest_widget_memory_save (conf, G_OBJECT(priv->main_paned),
				   "modest-main-paned");
	modest_widget_memory_save (conf, G_OBJECT(self->header_view), "header-view");
}


static void
on_connection_changed (TnyDevice *device, gboolean online, ModestMainWindow *self)
{
	GtkWidget *icon;
	const gchar *icon_name;
	ModestMainWindowPrivate *priv;
	
	g_return_if_fail (device);
	g_return_if_fail (self);

	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(self);

	icon_name = online ? GTK_STOCK_CONNECT : GTK_STOCK_DISCONNECT;
	icon      = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_BUTTON);

	/* Block handlers in order to avoid unnecessary calls */
	//g_signal_handler_block (G_OBJECT (priv->online_toggle), priv->toggle_button_signal);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(priv->online_toggle), online);
	//g_signal_handler_unblock (G_OBJECT (online_toggle), priv->toggle_button_signal);

	gtk_button_set_image (GTK_BUTTON(priv->online_toggle), icon);
	//statusbar_push (widget_factory, 0, online ? _("Modest went online") : _("Modest went offline"));
	
	/* If Modest has became online and the header view has a
	   header selected then show it */
	/* FIXME: there is a race condition if some account needs to
	   ask the user for a password */

/* 	if (online) { */
/* 		GtkTreeSelection *selected; */

/* 		selected = gtk_tree_view_get_selection (GTK_TREE_VIEW (header_view)); */
/* 		_modest_header_view_change_selection (selected, header_view); */
/* 	} */
}

void
on_online_toggle_toggled (GtkToggleButton *toggle, ModestMainWindow *self)
{
	gboolean online;
	TnyDevice *device;
	ModestMainWindowPrivate *priv;

	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(self);

	device = tny_account_store_get_device
		(TNY_ACCOUNT_STORE(modest_runtime_get_account_store()));

	online  = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(priv->online_toggle));

	if (online)
		tny_device_force_online (device);
	else
		tny_device_force_offline (device);

	g_object_unref (G_OBJECT (device));
}

static gboolean
on_delete_event (GtkWidget *widget, GdkEvent  *event, ModestMainWindow *self)
{
	save_sizes (self);
	return FALSE;
}

static void
on_destroy (GtkWidget *widget, GdkEvent  *event, ModestMainWindow *self)
{
	gtk_main_quit();
}



static void
connect_signals (ModestMainWindow *self)
{	
	ModestWindowPrivate *parent_priv;
	ModestMainWindowPrivate *priv;
	TnyDevice *device;
	ModestTnyAccountStore *account_store;
	
	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(self);
	parent_priv = MODEST_WINDOW_GET_PRIVATE(self);

	account_store = modest_runtime_get_account_store ();
	device        = tny_account_store_get_device(TNY_ACCOUNT_STORE(account_store));
	
	/* folder view */
	g_signal_connect (G_OBJECT(self->folder_view), "folder_selection_changed",
			  G_CALLBACK(modest_ui_actions_on_folder_selection_changed), self);
	g_signal_connect (G_OBJECT(self->folder_view), "folder_moved",
			  G_CALLBACK(modest_ui_actions_on_folder_moved), NULL);
	g_signal_connect (G_OBJECT(self->folder_view), "button-press-event",
			  G_CALLBACK (on_folder_view_button_press_event),self);
	g_signal_connect (self->folder_view,"popup-menu",
			  G_CALLBACK (on_folder_view_button_press_event),self);

	/* header view */
	g_signal_connect (G_OBJECT(self->header_view), "status_update",
			  G_CALLBACK(modest_ui_actions_on_header_status_update), self);
	g_signal_connect (G_OBJECT(self->header_view), "header_selected",
			  G_CALLBACK(modest_ui_actions_on_header_selected), self);
	g_signal_connect (G_OBJECT(self->header_view), "header_activated",
			  G_CALLBACK(modest_ui_actions_on_header_activated), self);
	g_signal_connect (G_OBJECT(self->header_view), "item_not_found",
			  G_CALLBACK(modest_ui_actions_on_item_not_found), self);
	g_signal_connect (G_OBJECT(self->header_view), "button-press-event",
			  G_CALLBACK (on_header_view_button_press_event), self);
	g_signal_connect (G_OBJECT(self->header_view),"popup-menu",
			  G_CALLBACK (on_header_view_button_press_event), self);
		
	/* msg preview */
	g_signal_connect (G_OBJECT(self->msg_preview), "link_clicked",
			  G_CALLBACK(modest_ui_actions_on_msg_link_clicked), self);
	g_signal_connect (G_OBJECT(self->msg_preview), "link_hover",
			  G_CALLBACK(modest_ui_actions_on_msg_link_hover), self);
	g_signal_connect (G_OBJECT(self->msg_preview), "attachment_clicked",
			  G_CALLBACK(modest_ui_actions_on_msg_attachment_clicked), self);

	/* Account store */
	g_signal_connect (G_OBJECT (modest_runtime_get_account_store()), "accounts_reloaded",
			  G_CALLBACK (modest_ui_actions_on_accounts_reloaded), self);
	
	/* Device */
	g_signal_connect (G_OBJECT(device), "connection_changed",
			  G_CALLBACK(on_connection_changed), self);
	g_signal_connect (G_OBJECT(priv->online_toggle), "toggled",
			  G_CALLBACK(on_online_toggle_toggled), NULL);
	
	/* window */
	g_signal_connect (G_OBJECT(self), "destroy", G_CALLBACK(on_destroy), NULL);
	g_signal_connect (G_OBJECT(self), "delete-event", G_CALLBACK(on_delete_event), self);
}


static GtkWidget*
wrapped_in_scrolled_window (GtkWidget *widget, gboolean needs_viewport)
{
	GtkWidget *win;

	win = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy
		(GTK_SCROLLED_WINDOW (win),GTK_POLICY_NEVER,
		 GTK_POLICY_AUTOMATIC);
	
	if (needs_viewport)
		gtk_scrolled_window_add_with_viewport
			(GTK_SCROLLED_WINDOW(win), widget);
	else
		gtk_container_add (GTK_CONTAINER(win),
				   widget);

	return win;
}




ModestWindow *
modest_main_window_new (void)
{
	GObject *obj;
	ModestMainWindow *self;
	ModestMainWindowPrivate *priv;
	ModestWindowPrivate *parent_priv;
	GtkWidget *main_vbox;
	GtkWidget *status_hbox;
	GtkWidget *header_win, *folder_win;
	GtkActionGroup *action_group;
	GError *error = NULL;
		
	obj  = g_object_new(MODEST_TYPE_MAIN_WINDOW, NULL);
	self = MODEST_MAIN_WINDOW(obj);
	
	priv = MODEST_MAIN_WINDOW_GET_PRIVATE(self);
	parent_priv = MODEST_WINDOW_GET_PRIVATE(self);
	
	/* ***************** */
	parent_priv->ui_manager = gtk_ui_manager_new();
	action_group = gtk_action_group_new ("ModestMainWindowActions");
	
	/* Add common actions */
	gtk_action_group_add_actions (action_group,
				      modest_action_entries,
				      G_N_ELEMENTS (modest_action_entries),
				      obj);

	gtk_ui_manager_insert_action_group (parent_priv->ui_manager, action_group, 0);
	g_object_unref (action_group);

	/* Load the UI definition */
	gtk_ui_manager_add_ui_from_file (parent_priv->ui_manager,
					 MODEST_UIDIR "modest-main-window-ui.xml", &error);
	if (error != NULL) {
		g_printerr ("modest: could not merge modest-main-window-ui.xml: %s", error->message);
		g_error_free (error);
		error = NULL;
	}
	/* *************** */

	/* Add accelerators */
	gtk_window_add_accel_group (GTK_WINDOW (obj), 
				    gtk_ui_manager_get_accel_group (parent_priv->ui_manager));

	/* Toolbar / Menubar */
	parent_priv->toolbar = gtk_ui_manager_get_widget (parent_priv->ui_manager, "/ToolBar");
	parent_priv->menubar = gtk_ui_manager_get_widget (parent_priv->ui_manager, "/MenuBar");

	gtk_toolbar_set_tooltips (GTK_TOOLBAR (parent_priv->toolbar), TRUE);	
	folder_win = wrapped_in_scrolled_window (GTK_WIDGET(self->folder_view), FALSE);
	header_win = wrapped_in_scrolled_window (GTK_WIDGET(self->header_view), FALSE);			   

	/* paned */
	priv->folder_paned = gtk_vpaned_new ();
	priv->msg_paned = gtk_vpaned_new ();
	priv->main_paned = gtk_hpaned_new ();
	gtk_paned_add1 (GTK_PANED(priv->main_paned), folder_win);
	gtk_paned_add2 (GTK_PANED(priv->main_paned), priv->msg_paned);
	gtk_paned_add1 (GTK_PANED(priv->msg_paned), header_win);
	gtk_paned_add2 (GTK_PANED(priv->msg_paned), GTK_WIDGET(self->msg_preview));

	gtk_widget_show (GTK_WIDGET(self->header_view));
	
	/* status bar / progress */
	status_hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(status_hbox), priv->folder_info_label, FALSE,FALSE, 6);
	gtk_box_pack_start (GTK_BOX(status_hbox), self->status_bar, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(status_hbox), self->progress_bar,FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(status_hbox), priv->online_toggle,FALSE, FALSE, 0);

	/* putting it all together... */
	main_vbox = gtk_vbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX(main_vbox), parent_priv->menubar, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(main_vbox), parent_priv->toolbar, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(main_vbox), priv->main_paned, TRUE, TRUE,0);
	gtk_box_pack_start (GTK_BOX(main_vbox), status_hbox, FALSE, FALSE, 0);
	
	gtk_container_add (GTK_CONTAINER(obj), main_vbox);
	restore_sizes (MODEST_MAIN_WINDOW(obj));	

	gtk_window_set_title (GTK_WINDOW(obj), _("Modest"));
	gtk_window_set_icon_from_file  (GTK_WINDOW(obj), MODEST_APP_ICON, NULL);	
	gtk_widget_show_all (main_vbox);
	
	/* Init toggle in correct state */
	//modest_ui_actions_on_connection_changed (device, tny_device_is_online (device), self);

	connect_signals (MODEST_MAIN_WINDOW(obj));
	return (ModestWindow *) obj;
}

static gboolean 
on_header_view_button_press_event (ModestHeaderView *header_view,
				   GdkEventButton   *event,
				   ModestMainWindow *self)
{
	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
		GtkWidget *menu;
		ModestWindowPrivate *parent_priv;
	
		parent_priv = MODEST_WINDOW_GET_PRIVATE (self);
		menu = gtk_ui_manager_get_widget (parent_priv->ui_manager, "/HeaderViewContextMenu");

		return show_context_popup_menu (self,
						GTK_TREE_VIEW (header_view), 
						event, 
						menu);
        }

        return FALSE;
}

static gboolean 
on_folder_view_button_press_event (ModestFolderView *folder_view,
				   GdkEventButton   *event,
				   ModestMainWindow *self)
{
	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
		GtkWidget *menu;
		ModestWindowPrivate *parent_priv;
	
		parent_priv = MODEST_WINDOW_GET_PRIVATE (self);
		menu = gtk_ui_manager_get_widget (parent_priv->ui_manager, "/FolderViewContextMenu");

		return show_context_popup_menu (self,
						GTK_TREE_VIEW (folder_view), 
						event, 
						menu);
        }

        return FALSE;
}

static gboolean 
show_context_popup_menu (ModestMainWindow *window,
			 GtkTreeView *tree_view,
			 GdkEventButton   *event,			 
			 GtkWidget *menu)
{
	g_return_val_if_fail (menu, FALSE);

        if (event != NULL) {
		/* Ensure that the header is selected */
		GtkTreeSelection *selection;

		selection = gtk_tree_view_get_selection (tree_view);
	
		if (gtk_tree_selection_count_selected_rows (selection) <= 1) {
			GtkTreePath *path;
		
			/* Get tree path for row that was clicked */
			if (gtk_tree_view_get_path_at_pos (tree_view,
							   (gint) event->x, 
							   (gint) event->y,
							   &path, 
							   NULL, NULL, NULL)) {
				gtk_tree_selection_unselect_all (selection);
				gtk_tree_selection_select_path (selection, path);
				gtk_tree_path_free (path);
			}
		}

		/* Show popup */
		if (gtk_tree_selection_count_selected_rows(selection) == 1)
			gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
					NULL, NULL,
					event->button, event->time);
	}
	return TRUE;
}
